#ifndef _PROPLASER_H_
#define _PROPLASER_H_

#include <propeller.h>
#include <simpletools.h>
#include <fdserial.h>
#include "dc_motor.h"

extern fdserial * txt;
extern DCMotor laser;


typedef uint8_t byte;

const int M1_Step = 1<<26;
const int M1_Dir  = 1<<25;

const int M2_Step = 1<<2;
const int M2_Dir = 1<<3;

const int XLimit_Min = 1<<18;
const int XLimit_Max = 1<<19;

const int YLimit_Min = 1<<10;
const int YLimit_Max = 1<<11;

const int Servo_A = 1<<12;
const int Servo_B = 1<<13;



struct FloatPoint {
  float x;
  float y;
  float z;
};

struct LongPoint {
  long  x;
  long  y;
  long  z;
};

struct GCodeParams {
  float x, y, z;
  float i, j, k;
  float f;

  byte paramValid;    // X = bit0, Y = bit1, etc...
};


// Stored as thousandths
struct IGCodeParams {
  long  x, y, z;
  long  i, j, k;
  long  f, p;

  byte paramValid;    // X = bit0, Y = bit1, etc...
};


extern LongPoint current_funits; // current units in fixed point
extern LongPoint target_funits;
extern LongPoint delta_funits;


extern LongPoint current_steps;
extern LongPoint target_steps;
extern LongPoint delta_steps;


extern float X_STEPS_PER_INCH;
extern float X_STEPS_PER_MM;
extern int X_MOTOR_STEPS;

extern float Y_STEPS_PER_INCH;
extern float Y_STEPS_PER_MM;
extern int Y_MOTOR_STEPS;

extern float Z_STEPS_PER_INCH;
extern float Z_STEPS_PER_MM;
extern int Z_MOTOR_STEPS;

//our maximum feedrates
extern int FAST_XY_FEEDRATE;
extern int FAST_Z_FEEDRATE;


extern byte X_STEP_PIN;
extern byte X_DIR_PIN;
extern byte X_MIN_PIN;
extern byte X_MAX_PIN;
extern byte Y_STEP_PIN;
extern byte Y_DIR_PIN;
extern byte Y_MIN_PIN;
extern byte Y_MAX_PIN;
extern byte Z_STEP_PIN;
extern byte Z_DIR_PIN;
extern byte Z_MIN_PIN;
extern byte Z_MAX_PIN;
extern byte Z_ENABLE_SERVO;
extern byte Z_ENABLE_LASER;

extern byte x_direction;
extern byte y_direction;
extern byte z_direction;

extern uint32_t XSTEP_MASK;
extern uint32_t YSTEP_MASK;
extern uint32_t XDIR_MASK;
extern uint32_t YDIR_MASK;
extern uint32_t XMIN_MASK;
extern uint32_t YMIN_MASK;
extern uint32_t XMAX_MASK;
extern uint32_t YMAX_MASK;

extern float x_units;
extern float y_units;
extern float z_units;

extern long  x_funits;
extern long  y_funits;
extern long  z_funits;


// Set to one if sensor outputs inverting (ie: 1 means open, 0 means closed)
// RepRap opto endstops are *not* inverting.
extern int SENSORS_INVERTING;
extern byte serial_count;


void dda_move(long micro_delay , byte rapid);
void dda_circle( long micro_delay , long xc, long yc, int8_t dir );
bool can_step(uint32_t min_pin, uint32_t max_pin, /*long current, long target, */ byte direction);

long to_fsteps(long steps_per_unit, long units);
void set_ftarget(long x, long y, long z);  // Set target location in fixed-point units
void set_fposition(long x, long y, long z);
void calculate_fdeltas();
long calculate_feedrate_delay(float feedrate);
long calculate_feedrate_delay_circle(float feedrate , float distance);
long getMaxSpeed();

#define OUTPUT 1
#define INPUT  0

#define HIGH   1
#define LOW    0

inline void pinMode(char pin, char mode) { if(mode == OUTPUT) DIRA |= (1<<pin); else DIRA &= ~(1<<pin); }
inline void digitalWrite(char pin, char val) { if(val) OUTA |= (1<<pin); else OUTA &= ~(1<<pin); }

inline int digitalPinToBitMask(char pin) {return 1<<pin;}

inline int max(int a, int b ) {return a>=b ? a : b;}
inline int min(int a, int b ) {return a<=b ? a : b;}

#endif
