
// This code has been rewritten for speed by Jason Dorie

//#include "Makeblock.h"
//#include <Servo.h>

#include "proplaser.h"

// define the parameters of our machine.
float X_STEPS_PER_INCH = 48;
float X_STEPS_PER_MM = 40;
int X_MOTOR_STEPS   = 100;

float Y_STEPS_PER_INCH = 48;
float Y_STEPS_PER_MM  = 40;
int Y_MOTOR_STEPS   = 100;

float Z_STEPS_PER_INCH = 48;
float Z_STEPS_PER_MM   = 40;
int Z_MOTOR_STEPS    = 100;

//our maximum feedrates
int FAST_XY_FEEDRATE = 5000;
int FAST_Z_FEEDRATE = 5000;

// Set to one if sensor outputs inverting (ie: 1 means open, 0 means closed)
// RepRap opto endstops are *not* inverting.
int SENSORS_INVERTING = 1;

DCMotor laser(8);


/****************************************************************************************
  digital i/o pin assignment

  this uses the undocumented feature of Arduino - pins 14-19 correspond to analog 0-5
****************************************************************************************/

byte X_STEP_PIN = 26;
byte X_DIR_PIN = 25;
byte X_MIN_PIN = 17;
byte X_MAX_PIN = 16;

byte Y_STEP_PIN = 2;
byte Y_DIR_PIN = 3;
byte Y_MIN_PIN = 13;
byte Y_MAX_PIN = 12;

uint32_t XSTEP_MASK = 1<<X_STEP_PIN;
uint32_t YSTEP_MASK = 1<<Y_STEP_PIN;
uint32_t XDIR_MASK = 1<<X_DIR_PIN;
uint32_t YDIR_MASK = 1<<Y_DIR_PIN;
uint32_t XMIN_MASK = 0;
uint32_t YMIN_MASK = 0;
uint32_t XMAX_MASK = 0;
uint32_t YMAX_MASK = 0;

byte Z_STEP_PIN = 15;
byte Z_DIR_PIN = -1;
byte Z_MIN_PIN = -1;
byte Z_MAX_PIN = -1;
byte Z_ENABLE_SERVO = 0;
byte Z_ENABLE_LASER = 1;

#define COMMAND_SIZE 128

char commands[COMMAND_SIZE];
byte serial_count;
int no_data = 0;

//Servo servo;

fdserial * txt;


void init_process_string(void);
void init_steppers(void);
void process_string(const char instruction[], byte size);

int currentPosServo = 90;
int targetPosServo = 90;
bool comment = false;

void setup()
{
  set_pause_dt( CLKFREQ/1000000 );

  //Do startup stuff here
  txt = fdserial_open( 31, 30, 0, 115200 );

  if (Z_ENABLE_SERVO == 1) {
    //servo.attach(Z_STEP_PIN);
  }

  //other initialization.
  init_process_string();
  init_steppers();
  process_string("G90", 3); //Absolute Position

  writeStr(txt, "start\n");

  laser.run(0);
}

void loop()
{
  int c;

  //read in characters if we got them.
  c = fdserial_rxCheck(txt);
  if( c >= 0 )
  {
    no_data = 0;
    //newlines are ends of commands.
    if (c != '\n')
    {
      if (c == 0x18) {
        writeStr(txt, "Grbl 1.0\n");
      }
      else {
        if (c == '(') {
          comment = true;
        }
        // If we're not in comment mode, add it to our array.
        if (!comment)
        {
          commands[serial_count] = c;
          serial_count++;
        }
        if (c == ')') {
          comment = false; // End of comment - start listening again
        }
      }
    }
    else if(serial_count) // got a command?
    {
      //process our command!
      process_string(commands, serial_count);

      //clear command.
      init_process_string();
    }
  }
  else
  {
    //if theres a pause or we got a real command, do it
    if( serial_count && (c == '\n' || no_data > 100) )
    {
      //process our command!
      process_string(commands, serial_count);

      //clear command.
      init_process_string();
    }
    else {
      no_data++;
      pause(50);
    }

    //no data?  turn off steppers
    if (no_data > 2000) {
      //disable_steppers();
      //laser.run(0);
    }
  }
}
