
#include "proplaser.h"


// This code has been rewritten for speed by Jason Dorie

long abs( long v ) {
  return (v>=0) ? v : -v;
}

void init_steppers_internal();


void init_steppers()
{
	//init our points.
  current_funits.x = 0;
  current_funits.y = 0;
  current_funits.z = 0;
  target_funits.x = 0;
  target_funits.y = 0;
  target_funits.z = 0;

	//figure our stuff.
  calculate_fdeltas();

  pinMode(X_STEP_PIN, OUTPUT);
  pinMode(X_DIR_PIN, OUTPUT);
  pinMode(X_MIN_PIN, INPUT);
  pinMode(X_MAX_PIN, INPUT);
	
  pinMode(Y_STEP_PIN, OUTPUT);
  pinMode(Y_DIR_PIN, OUTPUT);
  pinMode(Y_MIN_PIN, INPUT);
  pinMode(Y_MAX_PIN, INPUT);

  XMIN_MASK = digitalPinToBitMask(X_MIN_PIN);
  XMAX_MASK = digitalPinToBitMask(X_MAX_PIN);

  YMIN_MASK = digitalPinToBitMask(Y_MIN_PIN);
  YMAX_MASK = digitalPinToBitMask(Y_MAX_PIN);

  //pinMode(Z_STEP_PIN, OUTPUT);
  //pinMode(Z_DIR_PIN, OUTPUT);
  //pinMode(Z_MIN_PIN, INPUT);
  //pinMode(Z_MAX_PIN, INPUT);
}


void set_laser_power(void)
{
  int lz = -(int)current_steps.z;
  if( lz < 0 ) lz = 0;
  if( lz > 255 ) lz = 255;
  laser.run(lz);

  //dprinti( txt, "LZ: %d\n", lz );
}


/*
// acceleration stepping curve
int steps_per_unit;
float units_per_step;
float max_speed = 1.0;  // units per sec
float max_squared = max_speed * max_speed;  // units per sec
float accel = 1.0;  // units/sec^2 accel
int C0;
float FreqFactor = 1000000.0 * 0.676 * 256.0; // delays will be one microsec per step


void init_accel_move( int _steps_per_unit )
{
  steps_per_unit = _steps_per_unit;
  units_per_step = 1.0 / (float)steps_per_unit;

  float tmp = (2.0 * units_per_step ) / accel;
  C0 = round( FreqFactor * sqrt(tmp) );

  int accel_steps_required = round( MaxSquared / (2.0 * Accel * units_per_step) );
}
*/

static void digiWrite( uint32_t Mask, bool State )
{
  if( State ) {
    OUTA |= Mask;
  }
  else {
    OUTA &= ~Mask;
  }
}

static void digiToggle( uint32_t Mask )
{
  OUTA |= Mask;
  OUTA &= ~Mask;
}


static bool digiRead( uint32_t Mask )
{
  return (INA & Mask) != 0;
}


const long accelSteps_max = 6000;     // estimated
const byte ddaLoopOverhead = 5;       // estimated microseconds per loop to subtract from delay amount



void dda_move(long micro_delay , byte rapid)
{
  //Serial.print( current_steps.x ); Serial.print(" " ); Serial.print( current_steps.y ); Serial.print(" " ); Serial.println( current_steps.z );
  //Serial.print( target_steps.x  ); Serial.print(" " ); Serial.print( target_steps.y  ); Serial.print(" " ); Serial.println( target_steps.z  );
  //Serial.print( delta_steps.x  ); Serial.print(" " ); Serial.print( delta_steps.y  ); Serial.print(" " ); Serial.println( delta_steps.z  );

  long dx = delta_steps.x;
  long dy = delta_steps.y;
  long dz = delta_steps.z;

	//figure out our deltas
	int max_delta = max(dx, dy);
	max_delta = max(dz, max_delta);

  if( dx == 0 && dy == 0 ) rapid = 0; // don't do acceleration for Z - no need

	//init stuff.
	long x_counter = -(max_delta>>1);
	long y_counter = -(max_delta>>1);
	long z_counter = -(max_delta>>1);

	//our step flags
	bool x_can_step = 0;
	bool y_can_step = 0;
	bool z_can_step = 0;

  long Cn, accelSteps, decelStart;
  if( rapid ) {
    Cn = 2000 << 8;  // initial delay (was 800 - estimated - larger = slower acceleration)
    micro_delay = Cn >> 8;
    accelSteps = accelSteps_max;
    if( accelSteps > (max_delta>>1)) accelSteps = (max_delta>>1) - 1;
    decelStart = max_delta - accelSteps;  // step to start decelerating
  }

  int8_t xd = x_direction ? 1 : -1;
  int8_t yd = y_direction ? 1 : -1;
  int8_t zd = z_direction ? 1 : -1;

  digiWrite( XDIR_MASK , x_direction );
  digiWrite( YDIR_MASK , y_direction );

  //do our DDA line!
  long curStep = 1; // used only during rapid moves, for accel/decel tracking
  while( curStep <= max_delta )
  {
    //x_can_step = can_step(XMIN_MASK, XMAX_MASK, /*current_steps.x, target_steps.x,*/ x_direction);
    //y_can_step = can_step(YMIN_MASK, YMAX_MASK, /*current_steps.y, target_steps.y,*/ y_direction);
    //z_can_step = current_steps.z != target_steps.z; // no limits in Z because the laser is clamped


    x_counter += dx;
    y_counter += dy;
    z_counter += dz;

    if(x_counter > 0)
    {
      digiToggle( XSTEP_MASK );

      x_counter -= max_delta;
      current_steps.x += xd;
    }

    if(y_counter > 0)
    {
      digiToggle( YSTEP_MASK );

      y_counter -= max_delta;
      current_steps.y += yd;
    }

    if(z_counter > 0)
    {
      //if(Z_ENABLE_SERVO==0 && Z_ENABLE_LASER==0){
        //  do_step(Z_STEP_PIN, Z_DIR_PIN, z_direction);
      //}
      z_counter -= max_delta;

      current_steps.z += zd;
      set_laser_power();
    }

    //wait for next step.
    if( micro_delay > 0 )
      pause(micro_delay);

    curStep++;
    if( rapid ) {
      if( curStep < accelSteps ) {
        Cn -= (Cn << 1) / ((curStep<<2) + 1);
        micro_delay = (Cn >> 8) - ddaLoopOverhead; // tuning value, approximate # usecs taken by the loop
      }
      else if( curStep > decelStart ) {
        long n = curStep - max_delta;
        Cn -= (Cn << 1) / ((n<<2) + 1);
        micro_delay = (Cn >> 8) - ddaLoopOverhead;
      }
      else {
        pause(ddaLoopOverhead);
      }
    }
	}

	//set our points to be the same
  current_funits = target_funits;
  current_steps = target_steps;
}

void get_circle_dir( int8_t f, int8_t d, int8_t a, int8_t b , int8_t & xd, int8_t & yd )
{
  int binrep = 0;
  xd = yd = 0;
  if(f) binrep = binrep + 4;
  if(a) binrep = binrep + 2;
  if(b) binrep = binrep + 1;
  if(d) binrep = 7 - binrep;  // opposite direction, reverse the cases

  switch(binrep)
  {
    case 0:  yd = -1;  break;
    case 1:  xd = -1;  break;
    case 2:  xd =  1;  break;
    case 3:  yd =  1;  break;
    case 4:  xd =  1;  break;
    case 5:  yd = -1;  break;
    case 6:  yd =  1;  break;
    case 7:  xd = -1;  break;
  }
}

void dda_circle( long micro_delay , long xc, long yc, int8_t dir )
{
  //Serial.print( current_steps.x );  Serial.print("  " );
  //Serial.print( current_steps.y );  Serial.print("  " );
  //Serial.print( xc );  Serial.print("  " );
  //Serial.print( yc );  Serial.println("  " );
  //Serial.print( target_steps.x );  Serial.print("  " );
  //Serial.print( target_steps.y );  Serial.print("  " );
  //Serial.print( dir );  Serial.println("  " );

  dir ^= 1;

  //init stuff
  int xdelt = current_steps.x - xc;
  int ydelt = current_steps.y - yc;
  int64_t x_counter = xdelt * xdelt;
  int64_t y_counter = ydelt * ydelt;
  int64_t radsq = x_counter + y_counter;

  int8_t xd, yd;

  do {
    // figure out which direction we have to go
    int64_t fxy = x_counter + y_counter - radsq;
    int8_t f = fxy >= 0;
    int8_t a = xdelt >= 0;
    int8_t b = ydelt >= 0;
    
    get_circle_dir( f, dir, a, b , xd, yd );

    x_direction = xd >= 0;
    y_direction = yd >= 0;

    //Serial.print( current_steps.x );  Serial.print( "  " );  Serial.print( xd );  Serial.print( "  " );
    //Serial.print( current_steps.y );  Serial.print( "  " );  Serial.println( yd );

    if( xd != 0 ) {
      digiWrite( XDIR_MASK , x_direction );
      digiToggle( XSTEP_MASK );

      current_steps.x += xd;

      if( xd > 0 ) {
        x_counter += (xdelt<<1) + 1;
      }        
      else {
        x_counter -= (xdelt<<1) - 1;
      }        

      xdelt += xd;
      //x_counter = xdelt * xdelt;
    }

    if( yd != 0 ) {
      digiWrite( YDIR_MASK , y_direction );
      digiToggle( YSTEP_MASK );

      current_steps.y += yd;

      if( yd > 0 ) {
        y_counter += (ydelt<<1) + 1;
      }
      else {
        y_counter -= (ydelt<<1) + 1;
      }                

      ydelt += yd;
      //y_counter = ydelt * ydelt;
    }

    //wait for next step.
    if( micro_delay > 0 )
      pause(micro_delay);

  } while( (current_steps.x != target_steps.x && current_steps.y != target_steps.y) ||
          abs(current_steps.x - target_steps.x) > 10 || abs(current_steps.y - target_steps.y) > 10 );


  // only one of these loops will actually run...
  if( current_steps.x != target_steps.x ) {
    if( current_steps.x > target_steps.x )
      xd = -1;
    else
      xd = 1;
    x_direction = xd > 0;

    while( current_steps.x != target_steps.x )
    {
      digiWrite( XDIR_MASK , x_direction );
      digiToggle( XSTEP_MASK );
      current_steps.x += xd;

      //wait for   next step.
      if( micro_delay > 0 )
        pause(micro_delay);
    }
  }

  if( current_steps.y != target_steps.y ) {
    if( current_steps.y > target_steps.y )
      yd = -1;
    else
      yd = 1;
    y_direction = yd > 0;

    while( current_steps.y != target_steps.y )
    {
      digiWrite( YDIR_MASK , y_direction );
      digiToggle( YSTEP_MASK );
      current_steps.y += yd;

      //wait for next step.
      if( micro_delay > 0 )
        pause(micro_delay);
    }
  }

  current_funits.x = target_funits.x;
  current_funits.y = target_funits.y;
}


bool read_switch(uint32_t pin)
{
	//dual read as crude debounce
  if(SENSORS_INVERTING)
	  return !digiRead(pin) && !digiRead(pin);  // using && here to give preference to LOW signal
  else
    return digiRead(pin) && digiRead(pin);    // using && here to give preference to LOW signal
}



bool can_step(uint32_t min_pin, uint32_t max_pin, /*long current, long target, */byte direction)
{
	//stop us if we're on target
	//if (target == current)
	//	return false;

  /*
	//stop us if we're at home and still going 
	else if (!direction && read_switch(min_pin))
		return false;
	//stop us if we're at max and still going
	else if (direction && read_switch(max_pin))
		return false;
  */
	//default to being able to step
	return true;
}


long to_fsteps(long steps_per_unit, long units)
{
  long intRes = steps_per_unit * (units >> 8);
  long fracRes = steps_per_unit * (units & ((1<<8)-1));

  intRes += fracRes >> 8;
  return intRes >> 10;
}


void set_ftarget(long x, long y, long z)  // Set target location in fixed-point units
{
  target_funits.x = x;
  target_funits.y = y;
  target_funits.z = z;
 
  calculate_fdeltas();
}

void set_fposition(long x, long y, long z)
{
  current_funits.x = x;
  current_funits.y = y;
  current_funits.z = z;

  calculate_fdeltas();
}

void calculate_fdeltas()
{
  //figure our deltas.
  delta_funits.x = abs(target_funits.x - current_funits.x);
  delta_funits.y = abs(target_funits.y - current_funits.y);
  delta_funits.z = abs(target_funits.z - current_funits.z);

  //set our steps current, target, and delta
  //current_steps.x = to_steps(x_funits, current_funits.x);
  //current_steps.y = to_steps(y_funits, current_funits.y);
  //current_steps.z = to_steps(z_funits, current_funits.z);

  target_steps.x = to_fsteps(x_funits, target_funits.x);
  target_steps.y = to_fsteps(y_funits, target_funits.y);
  target_steps.z = to_fsteps(z_funits, target_funits.z);

  delta_steps.x = abs(target_steps.x - current_steps.x);
  delta_steps.y = abs(target_steps.y - current_steps.y);
  delta_steps.z = abs(target_steps.z - current_steps.z);

  //what is our direction
  x_direction = (target_funits.x >= current_funits.x);
  y_direction = (target_funits.y >= current_funits.y);
  z_direction = (target_funits.z >= current_funits.z);

  //set our direction pins as well
  //digitalWrite(X_DIR_PIN, x_direction^1);
  //digitalWrite(Y_DIR_PIN, y_direction);
  //digitalWrite(Z_DIR_PIN, z_direction);
}

float prevRate = 0;
float steps_per_second = 0;
float micros_per_step = 0;


long calculate_feedrate_delay(float feedrate)
{
  if( delta_steps.x == 0 && delta_steps.y == 0 ) {
    return 0; // No delay required for Z axis solo movements
  }

  // feedrate probably doesn't change too often, so cache some calculation results
  if( prevRate != feedrate ) {
    prevRate = feedrate;
    steps_per_second = feedrate * (1.0/60.0) * x_units;
    micros_per_step = 1000000.0 / steps_per_second;  // convert to microseconds per steps
  }

  // hypotenuse approximation - ~5.7% max error in x/y, z doesn't contribute anything significant in our usage
  long hi = max( delta_steps.x , delta_steps.y );
  long lo = min( delta_steps.x , delta_steps.y );
  hi = max( hi, delta_steps.z );
  long stepDist = (float)(hi + lo/3);

	long master_steps;

	//find the dominant axis.
	if (delta_steps.x >= delta_steps.y)
	{
		if (delta_steps.z > delta_steps.x)
			master_steps = delta_steps.z;
		else
			master_steps = delta_steps.x;
	}
	else
	{
		if (delta_steps.z > delta_steps.y)
			master_steps = delta_steps.z;
		else
			master_steps = delta_steps.y;
	}

  long res = (long)((stepDist * micros_per_step) / master_steps);
  if( res < 0 ) res = 0;

  return res;
}

long calculate_feedrate_delay_circle(float feedrate , float distance)
{
  float master_steps = x_units * distance;

  //calculate delay between steps in microseconds.  this is sort of tricky, but not too bad.
  //the formula has been condensed to save space.  here it is in english:
  // distance / feedrate * 60000000.0 = move duration in microseconds
  // move duration / master_steps = time between steps for master axis.

  long res = (distance * 60000000.0) / (feedrate * master_steps);
  if( res < 0 ) res = 0;

  return res;
}

long getMaxSpeed()
{
	if (delta_steps.z > 0)
		return calculate_feedrate_delay(FAST_Z_FEEDRATE);
	else
		return calculate_feedrate_delay(FAST_XY_FEEDRATE);
}

